/*
* BlockIP2S.c -	a Shared Memory Handler to establish a shred memory segment for the BlockIP2 security
*		channel exit to spped up the channel limitter.
*
* --------------------------------------------------------------------------
* Description
*
* This module is invoked by BlockIP2.dll and shall be paced in the exit library specified for the queue
* manager. Default path is: c:\program files\websphere mq\exits.
*
* This version of the program is a development edition and is without any
* guarantee, and use of it is on your own risk.
*
* Author:
*	   Jrgen Pedersen (c)
*	   email: oz1ccg@mrmq.dk
*
* Copyright:
*	   This program may be used and changed, but may not be sold
*	   or incoperated in any softwarepackage without the authors written permission.
*	   I would like to receive a copy of the improved version, to gather ideas
*	   for further improvements for my exit suite.
*
* --------------------------------------------------------------------------
* History
*
* Written by Jrgen Pedersen								  - 20Mar06 DK	v1.00
*
* Changes
*	   None yet
* --------------------------------------------------------------------------
*/

/* --------------------------------------------------------------------------
* Build hints
*
* windows:        cl BlockIP2S.c /MT mqm.lib
*
*/


#include <windows.h>
#include <stdio.h>
#include <conio.h>

#define MQ_CHANNEL_NAME_LENGTH 20
#define LOGDEBUG 2
#define LOGDATA  1

/*
 * MQ headers
 */
#include <cmqc.h>
#include <cmqxc.h>

#define SHMSZ 2048
#define SHMMAX_CHL ((SHMSZ-51) / sizeof(chlstruc)) -1

typedef struct _shmem_channels
{
    int  iNextChl;
    char szChannelName[MQ_CHANNEL_NAME_LENGTH+1];
    int  iActiveChannelLimit;
    int  iCurrentActiveConnections;
    int  iMaxActiveConnections;
} chlstruc;

typedef struct _shmem_dataspace
{							 /* communication structure */
    int iPid;                            /* who created/owns the shmem */
    int iNumOfChannels;
    char Reserved[43];

    chlstruc chnl[SHMMAX_CHL];
} shmem_type;

typedef struct WORKDATA
{
    /* start inq chl */
    MQHCONN	hConn;
    MQLONG 	HobjStat;
    int      bLogSuppress;              /* Suppress logging Loginternal when =1 */
    char     szStatusQ[MQ_OBJECT_NAME_LENGTH+1];
    char     szClientQ[MQ_OBJECT_NAME_LENGTH+1];
    char     szQMgr[MQ_Q_MGR_NAME_LENGTH+1];
    char     szConName[48+1];
    long     lMaxInUseAttempts;         /* number of tries to open queue */
    long     lWaitInterval;
    char     szTmpBufff[256];
#ifdef WIN32
    HANDLE   hMutexShm;
#endif
} WS;


void    LogData(int iMode, char * szData, WS *pWS);
long	connectToQMgr(WS *pWS);
long	DisConnectFromQMgr(WS *pWS);
long	openStatusQ(WS *pWS,long lOpenType, MQHOBJ* arg_pHobj);
long	closeStatusQ(WS *pWS);
long 	WaitUntilQmgrStop(WS *pWS);
void SendLogMessage (char *szData);

int main(int argc, char **argv)
{
    HANDLE hMapFile;
    LPCTSTR pBuf;

    long         lRc;

    WS          *pWS;         /* Working storage */
    WS           ws;

    HANDLE       semaphore;
    int          nRet;
    DWORD        retVal;
    DWORD        eNum;
    TCHAR        sysMsg[256];

    SECURITY_ATTRIBUTES		sec_attr;


    char szMutexKey[256];
    char szSemKey[256];
    shmem_type * shm = (shmem_type *) malloc(sizeof(shmem_type));

    pWS=&ws;

    strcpy(pWS->szStatusQ,"SYSTEM.DEFAULT.LOCAL.QUEUE");

    strcpy(szMutexKey,"Global\\BlockIP2-");
    strcat(szMutexKey,argv[1]);
    strcat(szMutexKey,"-shm");

    hMapFile = CreateFileMapping(
                   INVALID_HANDLE_VALUE,    // use paging file
                   NULL,                    // default security
                   PAGE_READWRITE,          // read/write access
                   0,                       // max. object size
                   SHMSZ,                   // buffer size
                   szMutexKey);             // name of mapping object

    if (hMapFile == NULL || hMapFile == INVALID_HANDLE_VALUE)
    {
        printf("Could not create file mapping object (%d).\n",
               GetLastError());
        return 8;
    }
    pBuf = (LPTSTR) MapViewOfFile(hMapFile,   // handle to map object
                                  FILE_MAP_ALL_ACCESS, // read/write permission
                                  0,
                                  0,
                                  SHMSZ);

    if (pBuf == NULL)
    {
        printf("Could not map view of file (%d).\n",
               GetLastError());
        return 8;
    }

    shm->iNumOfChannels=0;  /* No channes yet */


    /* establish semaphore */
    strcpy(szSemKey,"Global\\BlockIP2S-");
    strcat(szSemKey,argv[1]);
    strcat(szSemKey,"-ctrl");

    sprintf(pWS->szTmpBufff,"Semaphore key is [%s].\n",  szSemKey);
    LogData(LOGDEBUG, pWS->szTmpBufff, pWS);

    sec_attr.nLength              = sizeof( SECURITY_ATTRIBUTES );
    sec_attr.lpSecurityDescriptor = NULL;
    sec_attr.bInheritHandle       = TRUE;

    semaphore = CreateSemaphore( &sec_attr, 1, 65536, szSemKey );

    if( semaphore == (HANDLE) NULL ) {
        eNum = GetLastError( );
        FormatMessage( FORMAT_MESSAGE_FROM_SYSTEM |
                       FORMAT_MESSAGE_IGNORE_INSERTS,
                       NULL, eNum,
                       MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
                       sysMsg, 256, NULL );

        sprintf(pWS->szTmpBufff,"CreateSemaphore failed (%d) %s\n",   eNum, sysMsg);
        LogData(LOGDEBUG, pWS->szTmpBufff, pWS);
        return -1;
    }

    sprintf(pWS->szTmpBufff,"Semaphore created. (%lx)", nRet );
    LogData(LOGDEBUG, pWS->szTmpBufff, pWS);


    if( ! ReleaseSemaphore(semaphore, 1, NULL) ) {
        eNum = GetLastError( );
        FormatMessage( FORMAT_MESSAGE_FROM_SYSTEM |
                       FORMAT_MESSAGE_IGNORE_INSERTS,
                       NULL, eNum,
                       MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
                       sysMsg, 256, NULL );

        sprintf(pWS->szTmpBufff,"ReleaseSemaphore failed (%d) %s\n",   eNum, sysMsg);
        LogData(LOGDEBUG, pWS->szTmpBufff, pWS);
        return -1;
    }

    strcpy(pWS->szQMgr,argv[1]);

    if (connectToQMgr(pWS) != 0)
    {
        LogData(LOGDATA,"MQ Connect was not successfull, program terminated.",pWS);
        CloseHandle (semaphore);
        return 8;
    }

    lRc = openStatusQ(pWS,MQOO_BROWSE,&(pWS->HobjStat));
    if (lRc != MQRC_NONE)
    {
        printf("MQ open on %s was not successfull rc %ld,\nprogram terminated.\n",pWS->szStatusQ, lRc);
        CloseHandle (semaphore);
        return 8;
    }

    sprintf(pWS->szTmpBufff,"Queue %s opened hobj %ld .\n",pWS->szStatusQ, pWS->HobjStat);
    LogData(LOGDEBUG, pWS->szTmpBufff, pWS);

    CloseHandle (semaphore);
    LogData(LOGDEBUG,"Semaphore closed.",pWS);

    WaitUntilQmgrStop(pWS);

    //   getch();

    if (closeStatusQ(pWS) != MQRC_NONE)
    {
    }


    if (DisConnectFromQMgr(pWS) != MQRC_NONE)
    {
    }

    UnmapViewOfFile(pBuf);

    CloseHandle(hMapFile);
}


/*
**********************************************************************************************************
* Connect
*
**********************************************************************************************************
*/
long	connectToQMgr(WS *pWS) {
    long	lRc = 0;

    MQLONG	reason;
    MQLONG	compCode;

    sprintf(pWS->szTmpBufff,"Connect to default Qmgr %s.",pWS->szQMgr);
    LogData(LOGDEBUG, pWS->szTmpBufff, pWS);

    MQCONN(pWS->szQMgr,
           &pWS->hConn,
           &compCode,
           &reason);

    sprintf(pWS->szTmpBufff,"Debug connecting to %s. Reason=%ld hConn=%ld",pWS->szQMgr,reason,pWS->hConn);
    LogData(LOGDEBUG, pWS->szTmpBufff, pWS);

    if (reason == MQRC_ALREADY_CONNECTED || reason == MQRC_NONE){
    }else{
        lRc = 4;
        sprintf(pWS->szTmpBufff,"Error connecting to %s. Reason=%ld",pWS->szQMgr,reason);
        LogData(LOGDATA, pWS->szTmpBufff, pWS);
    }

    return lRc;
}


/*
**********************************************************************************************************
* Disconnect
*
**********************************************************************************************************
*/
long	DisConnectFromQMgr(WS *pWS){
    long	lRc = 0;

    MQLONG	reason;
    MQLONG	compCode;

    sprintf(pWS->szTmpBufff,"Dis connect to Qmgr %s.",pWS->szQMgr);
    LogData(LOGDEBUG, pWS->szTmpBufff, pWS);

    MQDISC(&pWS->hConn,
           &compCode,
           &reason);

    if (reason == MQRC_NONE){
    }else{
        lRc = 4;
        sprintf(pWS->szTmpBufff,"Error disconnecting from %s. Reason=%ld",pWS->szQMgr,reason);
        LogData(LOGDATA, pWS->szTmpBufff, pWS);
    }

    return lRc;
}

/*
**********************************************************************************************************
* Open Status queue.
**********************************************************************************************************
*/
long	openStatusQ(WS *pWS,long lOpenType, MQHOBJ* arg_pHobj) {

    long	lRc = 0;

    MQLONG reasonCode;
    MQLONG completionCode;
    MQLONG O_options;              /* MQOPEN options                */
    MQOD	od = {MQOD_DEFAULT};    /* Object Descriptor             */
    long lRetryCount;

    lRetryCount = pWS->lMaxInUseAttempts;

    LogData(LOGDEBUG,"In openStatusQ", pWS);

    /* Determine open options. */
    O_options = (lOpenType == MQOO_BROWSE)?MQOO_BROWSE|MQOO_FAIL_IF_QUIESCING:
                MQOO_INPUT_EXCLUSIVE | MQOO_OUTPUT|MQOO_BROWSE|MQOO_FAIL_IF_QUIESCING;

    sprintf(pWS->szTmpBufff,"Before MQOPEN status Q [%s] oo[%ld]", pWS->szStatusQ, O_options);
    LogData(LOGDEBUG, pWS->szTmpBufff, pWS);

    strcpy(od.ObjectName,pWS->szStatusQ);
    do {
        MQOPEN(pWS->hConn,        /* connection handle            */
               &od,               /* object descriptor for queue  */
               O_options,         /* open options                 */
               arg_pHobj,         /* object handle                */
               &completionCode,   /* MQOPEN completion code       */
               &reasonCode);      /* reason code                  */
        sprintf(pWS->szTmpBufff,"MQOPEN status Q ended with reason code %d on %s hConn=%ld", reasonCode,pWS->szStatusQ,pWS->hConn);
        LogData(LOGDEBUG, pWS->szTmpBufff, pWS);
        lRetryCount--;
    } while (reasonCode==MQRC_OBJECT_IN_USE && lRetryCount > 0);

    /* report reason, if any; stop if failed      */
    if (reasonCode != MQRC_NONE)
    {
        sprintf(pWS->szTmpBufff,"MQOPEN status Q ended with reason code %d on %s hConn=%ld", reasonCode,pWS->szStatusQ,pWS->hConn);
        LogData(LOGDATA, pWS->szTmpBufff, pWS);
        lRc = reasonCode ;
    }
    LogData(LOGDEBUG,"In leave openStatusQ", pWS);

    return lRc;
}

/*
**********************************************************************************************************
* Close status queue
*
**********************************************************************************************************
*/
long	closeStatusQ(WS *pWS) {
    long	lRc = 0;

    MQLONG	reason;
    MQLONG	compCode;
    MQLONG   C_options;              /* MQCLOSE options                */

    sprintf(pWS->szTmpBufff,"Close Status queue");
    LogData(LOGDEBUG, pWS->szTmpBufff, pWS);

    C_options = MQCO_NONE;   /* just close the queue */

    if (pWS->HobjStat != 0){
        MQCLOSE(pWS->hConn,
                &pWS->HobjStat,
                C_options,
                &compCode,
                &reason);

        pWS->HobjStat = 0;
        if (reason == MQRC_NONE){
        }else{
            lRc = 4;
            sprintf(pWS->szTmpBufff,"Error closing status queue %s. Reason=%ld",pWS->szStatusQ,reason);
            LogData(LOGDATA, pWS->szTmpBufff, pWS);
        }
    }

    pWS->HobjStat = 0;
    LogData(LOGDEBUG, "Close status Q done....", pWS);

    return lRc;
}



/***********************************************************************************************************
 *
 * pWS                work properties
 * lWaitInterval      waitinterval
 ***********************************************************************************************************/
long WaitUntilQmgrStop(WS *pWS)
{
    long	lRc = 0;

    char	szWork[40];

    MQLONG	reason;
    MQLONG	compCode;
    MQLONG	dataLength;

    MQGMO	gmo = {MQGMO_DEFAULT};
    MQMD	md = {MQMD_DEFAULT};

    LogData(LOGDEBUG,"In WaitUntilQmgrStop()", pWS);


    if (pWS->HobjStat == 0){
        /* Open the queue for browse */
        lRc = openStatusQ(pWS,MQOO_BROWSE,&(pWS->HobjStat));
    }

    sprintf(pWS->szTmpBufff,"lRc was %ld",lRc);
    LogData(LOGDATA, pWS->szTmpBufff, pWS);

    if (lRc == MQRC_NONE){
        /* Browse volatile data */
        strcpy((char*)md.MsgId,"BlockIP2S-Control");
        gmo.MatchOptions = MQMO_MATCH_MSG_ID;

        gmo.WaitInterval = -1;       /* wait forever */

        gmo.Options |= MQGMO_BROWSE_FIRST|MQGMO_ACCEPT_TRUNCATED_MSG|MQGMO_WAIT|MQGMO_FAIL_IF_QUIESCING;

        /* Browse. This will wait for the required number of millisecs (max) and then return */
        MQGET(pWS->hConn,
              pWS->HobjStat,
              &md,
              &gmo,
              sizeof(szWork),
              szWork,
              &dataLength,
              &compCode,
              &reason);
        if (reason != MQRC_NONE && reason != MQRC_NO_MSG_AVAILABLE){
            lRc = 4;
            sprintf(pWS->szTmpBufff,"Error on MQGET (browse wait for InUse processing) for queue %s. Reason=%ld",pWS->szStatusQ,reason);
            LogData(LOGDATA, pWS->szTmpBufff, pWS);
        }
    }

    LogData(LOGDEBUG,"Left WaitUntilQmgrStop()", pWS);

    return lRc;
}



/**
* LogData()
*
*
* Method to log data in a specific way
*
*	  iMode 0 = Normal 1= Debug String
*	 szData ASCII data to report
*
*
* @return void
*/
void LogData(int iMode, char * szData, WS *pWS)
{
    SendLogMessage(szData);
    printf("%s\n",szData);
}


/**
* SendLogMessage()
*
*
* Method to log data to EventList on windows
*
*	 szData ASCII data to report
*
*
* @return void
*
* added JHP 060214
*/

void SendLogMessage (char *szData)
{
    HANDLE  hSource;
    char    *szList[1];

    szList[0]=szData;
    hSource = RegisterEventSource(NULL, "WMQ BlockIP2");
    if (hSource != NULL)  {
        ReportEvent(hSource,EVENTLOG_INFORMATION_TYPE,
                    0,1,NULL,1,0,szList,NULL);
        DeregisterEventSource(hSource);
    }
}


